import typing
import logging

from . import types as boticord_types
from .http import HttpClient
from .autopost import AutoPost
from .exceptions import MeilisearchException

_logger = logging.getLogger("boticord")


class BoticordClient:
    """Represents a client that can be used to interact with the BotiCord API.

    Note:
        Remember that every http method can return an http exception.

    Args:
        token (:obj:`str`)
            Your bot's Boticord API Token.
        version (:obj:`int`)
            BotiCord API version (Default: 3)
    """

    __slots__ = ("http", "_autopost", "_token", "_meilisearch_api_key")

    http: HttpClient

    def __init__(self, token: str = None, version: int = 3):
        self._token = token
        self._meilisearch_api_key = None
        self._autopost: typing.Optional[AutoPost] = None
        self.http = HttpClient(token, version)

    async def get_bot_info(
        self, bot_id: typing.Union[str, int]
    ) -> boticord_types.ResourceBot:
        """Gets information about specified bot.

        Args:
            bot_id (Union[:obj:`str`, :obj:`int`])
                Id of the bot

        Returns:
            :obj:`~.types.ResourceBot`:
                ResourceBot object.
        """
        _logger.info("Requesting information about bot")

        response = await self.http.get_bot_info(bot_id)
        return boticord_types.ResourceBot.from_dict(response)

    async def post_bot_stats(
        self,
        bot_id: typing.Union[str, int],
        *,
        servers: int = 0,
        shards: int = 0,
        users: int = 0,
    ) -> boticord_types.ResourceBot:
        """Post Bot's stats.

        Args:
            bot_id (Union[:obj:`str`, :obj:`int`])
                Id of the bot to post stats of.
            servers ( :obj:`int` )
                Bot's servers count
            shards ( :obj:`int` )
                Bot's shards count
            users ( :obj:`int` )
                Bot's users count

        Returns:
            :obj:`~.types.ResourceBot`:
                ResourceBot object.
        """
        _logger.info("Posting bot stats")

        response = await self.http.post_bot_stats(
            bot_id, {"servers": servers, "shards": shards, "users": users}
        )
        return boticord_types.ResourceBot.from_dict(response)

    async def get_server_info(
        self, server_id: typing.Union[str, int]
    ) -> boticord_types.ResourceServer:
        """Gets information about specified server.

        Args:
            server_id (Union[:obj:`str`, :obj:`int`])
                Id of the server

        Returns:
            :obj:`~.types.ResourceServer`:
                ResourceServer object.
        """
        _logger.info("Requesting information about server")

        response = await self.http.get_server_info(server_id)
        return boticord_types.ResourceServer.from_dict(response)

    async def get_user_info(
        self, user_id: typing.Union[str, int]
    ) -> boticord_types.UserProfile:
        """Gets information about specified user.

        Args:
            user_id (Union[:obj:`str`, :obj:`int`])
                Id of the user

        Returns:
            :obj:`~.types.UserProfile`:
                UserProfile object.
        """
        _logger.info("Requesting information about user")

        response = await self.http.get_user_info(user_id)
        return boticord_types.UserProfile.from_dict(response)

    async def __search_for(self, index, data):
        """Search for something on BotiCord"""
        if self._meilisearch_api_key is None:
            token_response = await self.http.get_search_key()
            self._meilisearch_api_key = token_response["key"]

        try:
            response = await self.http.search_for(
                index, self._meilisearch_api_key, data
            )
        except MeilisearchException:
            token_response = await self.http.get_search_key()
            self._meilisearch_api_key = token_response["key"]

            response = await self.http.search_for(
                index, self._meilisearch_api_key, data
            )

        return response["hits"]

    async def search_for_bots(
        self, **kwargs
    ) -> typing.List[boticord_types.MeiliIndexedBot]:
        """Search for bots on BotiCord.

        Note:
            You can find every keyword argument `here <https://www.meilisearch.com/docs/reference/api/search#search-parameters>`_.

        Returns:
            List[:obj:`~.types.MeiliIndexedBot`]:
                List of found bots
        """
        _logger.info("Searching for bots on BotiCord")

        response = await self.__search_for("bots", kwargs)
        return [boticord_types.MeiliIndexedBot.from_dict(bot) for bot in response]

    async def search_for_servers(
        self, **kwargs
    ) -> typing.List[boticord_types.MeiliIndexedServer]:
        """Search for servers on BotiCord.

        Note:
            You can find every keyword argument `here <https://www.meilisearch.com/docs/reference/api/search#search-parameters>`_.

        Returns:
            List[:obj:`~.types.MeiliIndexedServer`]:
                List of found servers
        """
        _logger.info("Searching for servers on BotiCord")

        response = await self.__search_for("servers", kwargs)
        return [
            boticord_types.MeiliIndexedServer.from_dict(server) for server in response
        ]

    async def search_for_comments(
        self, **kwargs
    ) -> typing.List[boticord_types.MeiliIndexedComment]:
        """Search for comments on BotiCord.

        Note:
            You can find every keyword argument `here <https://www.meilisearch.com/docs/reference/api/search#search-parameters>`_.

        Returns:
            List[:obj:`~.types.MeiliIndexedComment`]:
                List of found comments
        """
        _logger.info("Searching for comments on BotiCord")

        response = await self.__search_for("comments", kwargs)
        return [
            boticord_types.MeiliIndexedComment.from_dict(comment)
            for comment in response
        ]

    def autopost(self) -> AutoPost:
        """Returns a helper instance for auto-posting.

        Returns:
            :obj:`~.autopost.AutoPost`: An instance of AutoPost.
        """
        if self._autopost is not None:
            return self._autopost

        self._autopost = AutoPost(self)
        return self._autopost



from enum import IntEnum


class BoticordException(Exception):
    """Base exception class for boticordpy.
    This could be caught to handle any exceptions thrown from this library.
    """


class InternalException(BoticordException):
    """Exception that's thrown when an local operation operation fails.

    Attributes
    ----------
    response:
        The meaning of the exception
    """

    def __init__(self, response):
        super().__init__(response)


class HTTPException(BoticordException):
    """Exception that's thrown when request to BotiCord API operation fails.

    Attributes
    ----------
    response:
        The response of the failed HTTP request.
    """

    def __init__(self, response):
        self.response = response

        fmt = f"{HTTPErrors(self.response['error']).name} (Status code: {StatusCodes(self.response['status']).name})"

        super().__init__(fmt)


class MeilisearchException(BoticordException):
    """Exception that's thrown when request to Meilisearch API operation fails.

    Attributes
    ----------
    response:
        The response of the failed HTTP request.
    """

    def __init__(self, response):
        self.response = response

        fmt = f"{self.response['code']} ({self.response['message']})"

        super().__init__(fmt)


class StatusCodes(IntEnum):
    """Status codes of response"""

    SERVER_ERROR = 500
    """Server Error (>500)"""

    TOO_MANY_REQUESTS = 429
    """Too Many Requests"""

    NOT_FOUND = 404
    """Requested resource was not found"""

    FORBIDDEN = 403
    """You don't have access to this resource"""

    UNAUTHORIZED = 401
    """Authorization is required to access this resource"""

    BAD_REQUEST = 400
    """Bad Request"""


class HTTPErrors(IntEnum):
    """Errors which BotiCord may return"""

    UNKNOWN_ERROR = 0
    """Unknown error"""

    INTERNAL_SERVER_ERROR = 1
    """Server error (>500)"""

    RATE_LIMITED = 2
    """Too many requests"""

    NOT_FOUND = 3
    """Not found"""

    FORBIDDEN = 4
    """Access denied"""

    BAD_REQUEST = 5
    """Bad request"""

    UNAUTHORIZED = 6
    """Unauthorized. Authorization required"""

    RPC_ERROR = 7
    """Server error (RPC)"""

    WS_ERROR = 8
    """Server error (WS)"""

    THIRD_PARTY_FAIL = 9
    """Third-party service error"""

    UNKNOWN_USER = 10
    """Unknown user"""

    SHORT_DOMAIN_TAKEN = 11
    """Short link already taken"""

    UNKNOWN_SHORT_DOMAIN = 12
    """Unknown short link"""

    UNKNOWN_LIBRARY = 13
    """Unknown library"""

    TOKEN_INVALID = 14
    """Invalid token"""

    UNKNOWN_RESOURCE = 15
    """Unknown resource"""

    UNKNOWN_TAG = 16
    """Unknown tag"""

    PERMISSION_DENIED = 17
    """Insufficient permissions"""

    UNKNOWN_COMMENT = 18
    """Unknown comment"""

    UNKNOWN_BOT = 19
    """Unknown bot"""

    UNKNOWN_SERVER = 20
    """Unknown server"""

    UNKNOWN_BADGE = 21
    """Unknown badge"""

    USER_ALREADY_HAS_A_BADGE = 22
    """User already has a badge"""

    INVALID_INVITE_CODE = 23
    """Invalid invite code"""

    SERVER_ALREADY_EXISTS = 24
    """Server already exists"""

    BOT_NOT_PRESENT_ON_QUEUE_SERVER = 25
    """Bot not present on queue server"""

    UNKNOWN_UP = 26
    """Unknown up"""

    TOO_MANY_UPS = 27
    """Too many ups"""

    INVALID_STATUS = 28
    """Invalid resource status"""

    UNKNOWN_REPORT = 29
    """Unknown report"""

    UNSUPPORTED_MEDIA_TYPE = 30
    """Unsupported media type. Should be one of"""

    UNKNOWN_APPLICATION = 31
    """Unknown application"""

    AUTOMATED_REQUESTS_NOT_ALLOWED = 32
    """Please confirm that you are not a robot by refreshing the page"""



import asyncio
import typing

import aiohttp

from . import exceptions


class HttpClient:
    """
    Represents an HTTP client sending HTTP requests to the BotiCord API.

    Args:
        token (:obj:`str`)
            Your bot's BotiCord API Token.

    Keyword Arguments:
        session: `aiohttp session`_
            The `aiohttp session`_ used for requests to the API.
        loop: `asyncio loop`
    """

    def __init__(self, auth_token: str = None, version: int = 3, **kwargs):
        self.token = auth_token
        self.API_URL = f"https://api.boticord.top/v{version}"

        loop = kwargs.get("loop") or asyncio.get_event_loop()

        self.session = kwargs.get("session") or aiohttp.ClientSession(loop=loop)

    async def make_request(
        self, method: str, endpoint: str, *, meilisearch_token: str = None, **kwargs
    ) -> dict:
        """Send requests to the API"""

        kwargs["headers"] = {"Content-Type": "application/json"}

        if self.token is not None:
            kwargs["headers"]["Authorization"] = self.token
        if meilisearch_token is not None:
            kwargs["headers"]["Authorization"] = f"Bearer {meilisearch_token}"

        url = f"{self.API_URL}{endpoint}"

        async with self.session.request(method, url, **kwargs) as response:
            data = await response.json()

            if (200, 201).__contains__(response.status):
                return data["result"] if not meilisearch_token else data
            else:
                if not meilisearch_token:
                    raise exceptions.HTTPException(
                        {"status": response.status, "error": data["errors"][0]["code"]}
                    )
                else:
                    raise exceptions.MeilisearchException(data)

    def get_bot_info(self, bot_id: typing.Union[str, int]):
        """Get information about the specified bot"""
        return self.make_request("GET", f"bots/{bot_id}")

    def post_bot_stats(self, bot_id: typing.Union[str, int], stats: dict):
        """Post bot's stats"""
        return self.make_request("POST", f"bots/{bot_id}/stats", json=stats)

    def get_server_info(self, server_id: typing.Union[str, int]):
        """Get information about specified server"""
        return self.make_request("GET", f"servers/{server_id}")

    def get_user_info(self, user_id: typing.Union[str, int]):
        """Get information about specified user"""
        return self.make_request("GET", f"users/{user_id}")

    def get_search_key(self):
        """Get API key for Meilisearch"""
        return self.make_request("GET", f"search-key")

    def search_for(self, index: str, api_key: str, data: dict):
        """Search for something on BotiCord."""
        return self.make_request(
            "POST",
            f"search/indexes/{index}/search",
            meilisearch_token=api_key,
            json=data,
        )



from datetime import datetime, timezone
from enum import IntEnum, Enum, EnumMeta
import copy
from dataclasses import _is_dataclass_instance, fields, dataclass
from typing import (
    Dict,
    Union,
    Generic,
    Tuple,
    TypeVar,
    get_origin,
    get_args,
    Optional,
    List,
)
from sys import modules
from itertools import chain

from typing_extensions import get_type_hints


KT = TypeVar("KT")
VT = TypeVar("VT")
T = TypeVar("T")


class Singleton(type):
    # Thanks to this stackoverflow answer (method 3):
    # https://stackoverflow.com/q/6760685/12668716
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class TypeCache(metaclass=Singleton):
    # Thanks to Pincer Devs. This class is from the Pincer Library.
    cache = {}

    def __init__(self):
        lcp = modules.copy()
        for module in lcp:
            if not module.startswith("melisa"):
                continue

            TypeCache.cache.update(lcp[module].__dict__)


def _asdict_ignore_none(obj: Generic[T]) -> Union[Tuple, Dict, T]:
    """
    Returns a dict from a dataclass that ignores
    all values that are None
    Modification of _asdict_inner from dataclasses
    Parameters
    ----------
    obj: Generic[T]
        The object to convert
    Returns
    -------
        A dict without None values
    """

    print(obj)

    if _is_dataclass_instance(obj):
        result = []
        for f in fields(obj):
            value = _asdict_ignore_none(getattr(obj, f.name))

            if isinstance(value, Enum):
                result.append((f.name, value.value))
            elif not f.name.startswith("_"):
                result.append((f.name, value))

        return dict(result)

    elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
        return type(obj)(*[_asdict_ignore_none(v) for v in obj])

    elif isinstance(obj, (list, tuple)):
        return type(obj)(_asdict_ignore_none(v) for v in obj)

    elif isinstance(obj, datetime):
        return str(round(obj.timestamp() * 1000))

    elif isinstance(obj, dict):
        return type(obj)(
            (_asdict_ignore_none(k), _asdict_ignore_none(v)) for k, v in obj.items()
        )
    else:
        return copy.deepcopy(obj)


class APIObjectBase:
    """
    Represents an object which has been fetched from the BotiCord API.
    """

    def __attr_convert(self, attr_value: Dict, attr_type: T) -> T:
        factory = attr_type

        # Always use `__factory__` over __init__
        if getattr(attr_type, "__factory__", None):
            factory = attr_type.__factory__

        if attr_value is None:
            return None

        if attr_type is not None and isinstance(attr_value, attr_type):
            return attr_value

        if isinstance(attr_value, dict):
            return factory(attr_value)

        return factory(attr_value)

    def __post_init__(self):
        TypeCache()

        attributes = chain.from_iterable(
            get_type_hints(cls, globalns=TypeCache.cache).items()
            for cls in chain(self.__class__.__bases__, (self,))
        )

        for attr, attr_type in attributes:
            # Ignore private attributes.
            if attr.startswith("_"):
                continue

            types = self.__get_types(attr, attr_type)

            types = tuple(filter(lambda tpe: tpe is not None, types))

            if not types:
                raise ValueError(
                    f"Attribute `{attr}` in `{type(self).__name__}` only "
                    "consisted of missing/optional type!"
                )

            specific_tp = types[0]

            attr_gotten = getattr(self, attr)

            if tp := get_origin(specific_tp):
                specific_tp = tp

            if isinstance(specific_tp, EnumMeta) and not attr_gotten:
                attr_value = None
            elif tp == list and attr_gotten and (classes := get_args(types[0])):
                attr_value = [
                    self.__attr_convert(attr_item, classes[0])
                    for attr_item in attr_gotten
                ]
            elif tp == dict and attr_gotten and (classes := get_args(types[0])):
                attr_value = {
                    key: self.__attr_convert(value, classes[1])
                    for key, value in attr_gotten.items()
                }
            else:
                attr_value = self.__attr_convert(attr_gotten, specific_tp)

            setattr(self, attr, attr_value)

    def __get_types(self, attr: str, arg_type: type) -> Tuple[type]:
        origin = get_origin(arg_type)

        if origin is Union:
            # Ahh yes, typing module has no type annotations for this...
            # noinspection PyTypeChecker
            args: Tuple[type] = get_args(arg_type)

            if 2 <= len(args) < 4:
                return args

            raise ValueError(
                f"Attribute `{attr}` in `{type(self).__name__}` has too many "
                f"or not enough arguments! (got {len(args)} expected 2-3)"
            )

        return (arg_type,)

    @classmethod
    def __factory__(cls: Generic[T], *args, **kwargs) -> T:
        return cls.from_dict(*args, **kwargs)

    def __repr__(self):
        attrs = ", ".join(
            f"{k}={v!r}"
            for k, v in self.__dict__.items()
            if v and not k.startswith("_")
        )

        return f"{type(self).__name__}({attrs})"

    def __str__(self):
        if _name := getattr(self, "__name__", None):
            return f"{_name} {self.__class__.__name__.lower()}"

        return super().__str__()

    def to_dict(self) -> Dict:
        """
        Transform the current object to a dictionary representation. Parameters that
        start with an underscore are not serialized.
        """
        return _asdict_ignore_none(self)


class BotLibrary(IntEnum):
    """The library that the bot is based on"""

    DISCORD4J = 1
    """Discord4j"""

    DISCORDCR = 2
    """Discordcr"""

    DISCORDGO = 3
    """DiscordGO"""

    DISCORDDOO = 4
    """Discordoo"""

    DSHARPPLUS = 5
    """DSharpPlus"""

    DISCORDJS = 6
    """Discord.js"""

    DISCORDNET = 7
    """Discord.Net"""

    DISCORDPY = 8
    """discord.py"""

    ERIS = 9
    """eris"""

    JAVACORD = 10
    """JavaCord"""

    JDA = 11
    """JDA"""

    OTHER = 12
    """Other"""

    NONE = 0
    """Bot's library doesn't specified"""


class ResourceStatus(IntEnum):
    """Status of the project on monitoring"""

    HIDDEN = 0
    """is hidden"""

    PUBLIC = 1
    """is public"""

    BANNED = 2
    """is banned"""

    PENDING = 3
    """is pending"""


class ServerTag(IntEnum):
    """Tags of the server"""

    SPEAKING = 130
    """Speaking"""

    FUN = 131
    """Fun"""

    GAMES = 132
    """Games"""

    CINEMA = 133
    """Cinema"""

    ANIME = 134
    """Anime"""

    ART = 135
    """Art"""

    CODING = 136
    """Coding"""

    MUSIC = 137
    """Music"""

    ADULT = 138
    """18+"""

    ROLEPLAY = 139
    """Role-Play"""

    HUMOUR = 140
    """Humour"""

    GENSHIN = 160
    """Genshin"""

    MINECRAFT = 161
    """Minecraft"""

    GTA = 162
    """GTA"""

    CS = 163
    """CS"""

    DOTA = 164
    """Dota"""

    AMONG_US = 165
    """Among Us"""

    FORTNITE = 166
    """Fortnite"""

    BRAWL_STARS = 167
    """Brawl Stars"""


class BotTag(IntEnum):
    """Tags of the bot"""

    MODERATION = 0
    """Moderation"""

    BOT = 1
    """Bot"""

    UTILITIES = 2
    """Utilities"""

    ENTERTAINMENT = 3
    """Entertainment"""

    MUSIC = 4
    """Music"""

    ECONOMY = 5
    """Economy"""

    LOGS = 6
    """Logs"""

    LEVELS = 7
    """Levels"""

    NSFW = 8
    """NSFW (18+)"""

    SETTINGS = 9
    """Settings"""

    ROLE_PLAY = 10
    """Role-Play"""

    MEMES = 11
    """Memes"""

    GAMES = 12
    """Games"""

    AI = 13
    """AI"""


@dataclass(repr=False)
class UserLinks(APIObjectBase):
    """Links of the userk"""

    vk: Optional[str]
    """vk.com"""

    telegram: Optional[str]
    """t.me"""

    donate: Optional[str]
    """Donate"""

    git: Optional[str]
    """Link to git of the user"""

    custom: Optional[str]
    """Custom link"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a UserLinks from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a UserLinks.
        """

        self: UserLinks = super().__new__(cls)
        data = data or {}

        self.vk = data.get("vk")
        self.telegram = data.get("telegram")
        self.donate = data.get("donate")
        self.git = data.get("git")
        self.custom = data.get("custom")

        return self


@dataclass(repr=False)
class UserBadge(APIObjectBase):
    """Information about user's profile badge"""

    id: int
    """Badge's ID"""

    name: str
    """Badge's name"""

    asset_url: str
    """Badge's icon URL"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a UserBadge from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a UserBadge.
        """
        self: UserBadge = super().__new__(cls)

        self.id = data["id"]
        self.name = data["name"]
        self.asset_url = data["assetURL"]

        return self


@dataclass(repr=False)
class ResourceUp(APIObjectBase):
    """Information about bump (bot/server)"""

    id: str
    """Bump's id"""

    expires: datetime
    """Expiration date. (ATTENTION! When using `to_dict()`, the data may not correspond to the actual data due to the peculiarities of the `datetime` module)"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a ResourceUp from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a ResourceUp.
        """

        self: ResourceUp = super().__new__(cls)

        self.id = data["id"]
        self.expires = datetime.fromtimestamp(
            int(int(data["expires"]) / 1000), tz=timezone.utc
        )

        return self


@dataclass(repr=False)
class ResourceRating(APIObjectBase):
    """Rating of bot/server"""

    count: int
    """Number of ratings"""

    rating: int
    """Rating (from 1 to 5)"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a ResourceRating from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a ResourceRating.
        """

        self: ResourceRating = super().__new__(cls)

        self.count = data["count"]
        self.rating = data["rating"]

        return self


@dataclass(repr=False)
class PartialUser(APIObjectBase):
    """Partial user from BotiCord."""

    username: str
    """Username"""

    discriminator: str
    """Discriminator"""

    avatar: Optional[str]
    """Avatar of the user"""

    id: str
    """Id of the user"""

    socials: UserLinks
    """Links of the user"""

    description: Optional[str]
    """Description of the user"""

    short_description: Optional[str]
    """Short description of the user"""

    status: Optional[str]
    """Status of the user"""

    short_domain: Optional[str]
    """Short domain"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a PartialUser from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a PartialUser.
        """

        self: cls = super().__new__(cls)

        self.username = data["username"]
        self.discriminator = data.get("discriminator")
        self.avatar = data.get("avatar")
        self.id = data["id"]
        self.socials = UserLinks.from_dict(data.get("socials", {}))
        self.description = data.get("description")
        self.short_description = data.get("shortDescription")
        self.status = data.get("status")
        self.short_domain = data.get("shortDomain")

        return self


@dataclass(repr=False)
class ResourceServer(APIObjectBase):
    """Information about server from BotiCord.

    .. warning::

        The result of the reverse conversion (`.to_dict()`) may not match the actual data.
    """

    id: str
    """Server's ID"""

    name: str
    """Server's name"""

    short_description: str
    """Server's short description"""

    description: str
    """Server's description"""

    avatar: Optional[str]
    """Server's avatar"""

    short_link: Optional[str]
    """Server's short link"""

    invite_link: str
    """Server's invite link"""

    premium_active: bool
    """Server's premium state"""

    premium_auto_fetch: Optional[bool]
    """Server's premium auto fetch state"""

    premium_banner_url: Optional[str]
    """Server's premium banner URL"""

    premium_splash_url: Optional[str]
    """Server's premium splash URL"""

    standart_banner_id: int
    """Server's standart banner ID"""

    owner: str
    """Server's owner ID"""

    status: ResourceStatus
    """Server's status"""

    ratings: List[ResourceRating]
    """Server's ratings"""

    created_date: datetime
    """Server's creation time"""

    members: Optional[int]
    """Server's members count"""

    website: Optional[str]
    """Server's website"""

    tags: List[ServerTag]
    """Server's tags"""

    moderators: List[PartialUser]
    """Server's moderators"""

    up_count: int
    """Server's up count"""

    ups: Optional[ResourceUp]
    """Server's ups"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a ResourceServer from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a ResourceServer."""

        self = super().__new__(cls)

        self.id = data.get("id")
        self.name = data.get("name")
        self.short_description = data.get("shortDescription")
        self.description = data.get("description")
        self.avatar = data.get("avatar")
        self.short_link = data.get("shortLink")
        self.invite_link = data.get("inviteLink")
        self.owner = data.get("owner")
        self.website = data.get("website")
        self.up_count = data.get("upCount")
        self.standart_banner_id = data.get("standartBannerID")

        self.premium_active = data["premium"].get("active")
        self.premium_splash_url = data["premium"].get("splashURL")
        self.premium_auto_fetch = data["premium"].get("autoFetch")
        self.premium_banner_url = data["premium"].get("bannerURL")

        self.status = ResourceStatus(data.get("status"))
        self.ratings = [
            ResourceRating.from_dict(rating) for rating in data.get("ratings", [])
        ]
        self.created_date = datetime.strptime(
            data.get("createdDate"), "%Y-%m-%dT%H:%M:%S.%f%z"
        )
        self.tags = [ServerTag(tag) for tag in data.get("tags", [])]
        self.ups = [ResourceUp.from_dict(up) for up in data.get("ups", [])]
        self.moderators = [
            PartialUser.from_dict(mod) for mod in data.get("moderators", [])
        ]

        self.members = data.get("memberCount")

        return self


@dataclass(repr=False)
class ResourceBot(APIObjectBase):
    """Bot published on BotiCord

    .. warning::

        The result of the reverse conversion (`.to_dict()`) may not match the actual data.
    """

    id: str
    """ID of the bot"""

    name: str
    """Name of the bot"""

    short_description: str
    """Short description of the bot"""

    description: str
    """Description of the bot"""

    avatar: Optional[str]
    """Avatar of the bot"""

    short_link: Optional[str]
    """Short link to the bot's page"""

    standart_banner_id: int
    """Server's standart banner ID"""

    invite_link: str
    """Invite link"""

    premium_active: bool
    """Is premium status active? (True/False)"""

    premium_splash_url: Optional[str]
    """Link to the splash"""

    premium_auto_fetch: Optional[bool]
    """Is auto-fetch enabled? (True/False)"""

    premium_banner_url: Optional[str]
    """Premium banner URL"""

    owner: str
    """Owner of the bot"""

    status: ResourceStatus
    """Status of the bot"""

    ratings: List[ResourceRating]
    """Bot's ratings"""

    prefix: str
    """Prefix of the bot"""

    discriminator: str
    """Bot's discriminator"""

    created_date: datetime
    """Date when the bot was published"""

    support_server_invite_link: Optional[str]
    """Link to the support server"""

    library: Optional[BotLibrary]
    """The library that the bot is based on"""

    guilds: Optional[int]
    """Number of guilds"""

    shards: Optional[int]
    """Number of shards"""

    members: Optional[int]
    """Number of members"""

    website: Optional[str]
    """Link to bot's website"""

    tags: List[BotTag]
    """List of bot tags"""

    up_count: int
    """Number of ups"""

    ups: List[ResourceUp]
    """List of bot's ups"""

    developers: List[PartialUser]
    """List of bot's developers"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a ResourceBot from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a ResourceBot.
        """

        self: ResourceBot = super().__new__(cls)

        self.id = data.get("id")
        self.name = data.get("name")
        self.short_description = data.get("shortDescription")
        self.description = data.get("description")
        self.avatar = data.get("avatar")
        self.short_link = data.get("shortLink")
        self.invite_link = data.get("inviteLink")
        self.owner = data.get("owner")
        self.prefix = data.get("prefix")
        self.discriminator = data.get("discriminator")
        self.support_server_invite_link = data.get("support_server_invite")
        self.website = data.get("website")
        self.up_count = data.get("upCount")
        self.standart_banner_id = data.get("standartBannerID")

        self.premium_active = data["premium"].get("active")
        self.premium_splash_url = data["premium"].get("splashURL")
        self.premium_auto_fetch = data["premium"].get("autoFetch")
        self.premium_banner_url = data["premium"].get("bannerURL")

        self.status = ResourceStatus(data.get("status"))
        self.ratings = [
            ResourceRating.from_dict(rating) for rating in data.get("ratings", [])
        ]
        self.created_date = datetime.strptime(
            data["createdDate"], "%Y-%m-%dT%H:%M:%S.%f%z"
        )
        self.library = (
            BotLibrary(data["library"]) if data.get("library") is not None else None
        )
        self.tags = [BotTag(tag) for tag in data.get("tags", [])]
        self.ups = [ResourceUp.from_dict(up) for up in data.get("ups", [])]
        self.developers = [
            PartialUser.from_dict(dev) for dev in data.get("developers", [])
        ]

        self.guilds = data.get("guilds")
        self.shards = data.get("shards")
        self.members = data.get("members")

        return self


@dataclass(repr=False)
class UserProfile(PartialUser):
    """Information about user's profile from BotiCord.'"""

    badges: List[UserBadge]
    """User's badges list."""

    bots: List[ResourceBot]
    """User's bots list"""

    servers: List[ResourceServer]
    """User's servers list"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a UserProfile from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a UserProfile."""

        self = super().from_dict(data)

        self.badges = [UserBadge.from_dict(badge) for badge in data.get("badges", [])]
        self.bots = [ResourceBot.from_dict(bot) for bot in data.get("bots", [])]
        self.servers = [
            ResourceServer.from_dict(server) for server in data.get("servers", [])
        ]

        return self


@dataclass(repr=False)
class MeiliIndexedBot(APIObjectBase):
    """Bot found on BotiCord

    .. warning::

        The result of the reverse conversion (`.to_dict()`) may not match the actual data.
    """

    id: str
    """ID of the bot"""

    name: str
    """Name of the bot"""

    short_description: str
    """Short description of the bot"""

    description: str
    """Description of the bot"""

    avatar: Optional[str]
    """Avatar of the bot"""

    invite: str
    """Invite link"""

    premium_active: bool
    """Is premium status active? (True/False)"""

    premium_banner: Optional[str]
    """Premium banner URL"""

    banner: int
    """Standart banner"""

    rating: int
    """Bot's rating"""

    discriminator: str
    """Bot's discriminator"""

    library: Optional[BotLibrary]
    """The library that the bot is based on"""

    guilds: Optional[int]
    """Number of guilds"""

    shards: Optional[int]
    """Number of shards"""

    members: Optional[int]
    """Number of members"""

    tags: List[BotTag]
    """List of bot tags"""

    ups: int
    """List of bot's ups"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a MeiliIndexedBot from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a MeiliIndexedBot.
        """

        self: MeiliIndexedBot = super().__new__(cls)

        self.id = data.get("id")
        self.name = data.get("name")
        self.short_description = data.get("shortDescription")
        self.description = data.get("description")
        self.avatar = data.get("avatar")
        self.invite = data.get("invite")
        self.discriminator = data.get("discriminator")
        self.ups = data.get("ups")
        self.rating = data.get("rating")
        self.banner = data.get("banner")

        self.premium_active = data.get("premiumActive")
        self.premium_banner = data.get("premiumBanner")

        self.library = (
            BotLibrary(data["library"]) if data.get("library") is not None else None
        )
        self.tags = [BotTag(tag) for tag in data.get("tags", [])]

        self.guilds = data.get("guilds")
        self.shards = data.get("shards")
        self.members = data.get("members")

        return self


@dataclass(repr=False)
class MeiliIndexedServer(APIObjectBase):
    """Server found on BotiCord

    .. warning::

        The result of the reverse conversion (`.to_dict()`) may not match the actual data.
    """

    id: str
    """ID of the server"""

    name: str
    """Name of the server"""

    short_description: str
    """Short description of the server"""

    description: str
    """Description of the server"""

    avatar: Optional[str]
    """Avatar of the server"""

    invite: str
    """Invite link"""

    premium_active: bool
    """Is premium status active? (True/False)"""

    premium_banner: Optional[str]
    """Premium banner URL"""

    banner: int
    """Standart banner"""

    discord_banner: Optional[str]
    """Discord banner URL"""

    rating: int
    """Server's rating"""

    members: Optional[int]
    """Number of members"""

    tags: List[ServerTag]
    """List of server tags"""

    ups: int
    """List of server's ups"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a MeiliIndexedServer from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a MeiliIndexedServer.
        """

        self: MeiliIndexedServer = super().__new__(cls)

        self.id = data.get("id")
        self.name = data.get("name")
        self.short_description = data.get("shortDescription")
        self.description = data.get("description")
        self.avatar = data.get("avatar")
        self.invite = data.get("invite")
        self.ups = data.get("ups")
        self.rating = data.get("rating")
        self.banner = data.get("banner")

        self.premium_active = data.get("premiumActive")
        self.premium_banner = data.get("premiumBanner")
        self.discord_banner = data.get("discordBanner")

        self.tags = [ServerTag(tag) for tag in data.get("tags", [])]

        self.members = data.get("members")

        return self


@dataclass(repr=False)
class MeiliIndexedComment(APIObjectBase):
    """Comment found on BotiCord"""

    id: str
    """ID of the comment"""

    author: str
    """Id of the author of the comment"""

    rating: int
    """Comment's rating"""

    content: str
    """Content of the comment"""

    resource: str
    """Id of the resource"""

    created: datetime
    """When the comment was created"""

    mod_reply: Optional[str]
    """Reply to the comment"""

    @classmethod
    def from_dict(cls, data: dict):
        """Generate a MeiliIndexedComment from the given data.

        Parameters
        ----------
        data: :class:`dict`
            The dictionary to convert into a MeiliIndexedComment.
        """

        self: MeiliIndexedComment = super().__new__(cls)

        self.id = data.get("id")
        self.rating = data.get("rating")
        self.author = data.get("author")
        self.content = data.get("content")
        self.resource = data.get("resource")
        self.mod_reply = data.get("modReply")
        self.created = datetime.utcfromtimestamp(data.get("created") / 1000)

        return self



# Copyright Marakarka (Viktor K) 2021 - Present
# Full MIT License can be found in `LICENSE.txt` at the project root.

import logging
import json
import asyncio
import typing

import aiohttp

_logger = logging.getLogger("boticord.websocket")


class BotiCordWebsocket:
    """Represents a client that can be used to interact with the BotiCord by websocket connection."""

    def __init__(self, token: str):
        self.__session = None
        self.loop = asyncio.get_event_loop()
        self.ws = None
        self._listeners = {}
        self.not_closed = True

        self._token = token

    def listener(self):
        """Decorator to set the listener.

        .. warning::

            Callback functions must be a **coroutine**. If they aren't, then you might get unexpected
            errors. In order to turn a function into a coroutine they must be ``async def``
            functions.

        For example:

        .. code-block:: python

            @websocket.listener()
            async def comment_removed(data):
                pass
        """

        def inner(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError(f"<{func.__qualname__}> must be a coroutine function")
            self._listeners[func.__qualname__] = func
            _logger.debug(f"Listener {func.__qualname__} added successfully!")
            return func

        return inner

    def register_listener(self, notification_type: str, callback: typing.Any):
        """Method to set the listener.

        Args:
            notify_type (:obj:`str`)
                Type of notification (Check reference page)
            callback (:obj:`function`)
                Coroutine Callback Function

        .. warning::

            Callback functions must be a **coroutine**. If they aren't, then you might get unexpected
            errors. In order to turn a function into a coroutine they must be ``async def``
            functions.
        """
        if not asyncio.iscoroutinefunction(callback):
            raise TypeError(f"<{callback.__qualname__}> must be a coroutine function")

        self._listeners[notification_type] = callback
        _logger.debug(f"Listener {callback.__qualname__} added successfully!")
        return self

    async def connect(self) -> None:
        """Connect to BotiCord."""
        try:
            self.__session = aiohttp.ClientSession()
            self.ws = await self.__session.ws_connect(
                "wss://gateway.boticord.top/websocket/",
                timeout=30.0,
            )

            _logger.info("Connected to BotiCord.")

            self.not_closed = True

            self.loop.create_task(self._receive())
            await self._send_identify()
        except Exception as exc:
            _logger.error("Connecting failed!")

            raise exc

    async def _send_identify(self) -> None:
        await self.ws.send_json({"event": "auth", "data": {"token": self._token}})

    async def _receive(self) -> None:
        while self.not_closed:
            async for msg in self.ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    await self._handle_data(msg.data)
                else:
                    raise RuntimeError

            close_code = self.ws.close_code

            if close_code is not None:
                await self._handle_close(close_code)

    async def _handle_data(self, data):
        data = json.loads(data)

        if data["event"] == "hello":
            _logger.info("Authorized successfully.")
            self.loop.create_task(self._send_ping())
        elif data["event"] == "notify":
            listener = self._listeners.get(data["data"]["type"])
            if listener:
                self.loop.create_task(listener(data["data"]))
        elif data["event"] == "pong":
            _logger.info("Received pong-response.")
            self.loop.create_task(self._send_ping())
        else:
            _logger.error("An error has occurred.")

    async def _handle_close(self, code: int) -> None:
        self.not_closed = False
        await self.__session.close()

        if code == 4000:
            _logger.info("Closed connection successfully.")
            return
        elif code == 1006:
            _logger.error("Token is invalid.")
            return

        _logger.info("Disconnected from BotiCord. Reconnecting...")

        await self.connect()

    async def _send_ping(self) -> None:
        if not self.ws.closed:
            await asyncio.sleep(45)
            await self.ws.send_json({"event": "ping"})

    async def close(self) -> None:
        """Close websocket connection with BotiCord"""
        if self.ws:
            self.not_closed = False
            await self.ws.close(code=4000)



import pathlib

from setuptools import setup, find_packages
import re

HERE = pathlib.Path(__file__).parent

with open("boticordpy/__init__.py") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(1)

if version.endswith(("a", "b", "rc")):
    # append version identifier based on commit count
    try:
        import subprocess

        p = subprocess.Popen(
            ["git", "rev-list", "--count", "HEAD"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = p.communicate()
        if out:
            version += out.decode("utf-8").strip()
        p = subprocess.Popen(
            ["git", "rev-parse", "--short", "HEAD"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = p.communicate()
        if out:
            version += "+g" + out.decode("utf-8").strip()
    except Exception:
        pass

README = (HERE / "README.md").read_text(encoding="utf8")

setup(
    name="boticordpy",
    project_urls={
        "Documentation": "https://py.boticord.top/en/stable",
        "Issue tracker": "https://github.com/boticord/boticordpy/issues",
    },
    packages=find_packages(),
    version=version,
    python_requires=">= 3.8",
    description="A Python wrapper for BotiCord API",
    long_description=README,
    long_description_content_type="text/markdown",
    include_package_data=True,
    url="https://github.com/boticord/boticordpy",
    author="Marakarka",
    author_email="support@kerdoku.top",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    install_requires=["aiohttp", "typing_extensions"],
)
